#include "retis_servomotor.h"

#ifdef __ERIKA__
#include "ee.h"
#else 
#error "RETIS_SERVOMOTOR: ERIKA OS is required by the motionboard." 
#endif

#ifndef RETIS_SERVOMOTOR_N
#define RETIS_SERVOMOTOR_N 2
#endif

#if RETIS_SERVOMOTOR_N > 2
#error "RETIS_SERVOMOTOR: BAD RETIS_SERVOMOTOR_N value, "
       "the max number of servomotors supported by the motionboard is 2."
#endif

#define SYSTEM_TICK 40

static struct {
	int16_t min_angle;
	int16_t max_angle;
	uint16_t ocr_at_zero;
	uint16_t ocr_unit;
} servo_stat[RETIS_SERVOMOTOR_N];

static uint8_t timer2_prescaler = -1;

/* OutputCompare PWM constants
* 
* pwm_timer2_period = (fcy/pre_scaler) x pwm_perdiod - 1
* 		fcy [MHz]
* 		pwm_perdiod [microseconds]
* 
* pwm_resolution = log_2((fcy/pre_scaler) x pwm_perdiod) [bits]
*  
*/

static int8_t timer2_init(uint32_t pwm_period)
{
	uint16_t ps, p;
	
	/* Calculate the best (minimum) pre-scaling factor */	
	p = (((SYSTEM_TICK * pwm_period) -1) / 0xFFFF) + 1; /* Ceiling */
	for (ps = 0; ps < 4; ps++) {
		if (ps < 3) {
			if (p <= (0x0001 << (ps * 3))) { 
				/* note: use shift for future division */
				p = (ps * 3);	
				break;
			}
		} else {
			if (p <= 256) { 
				/* note: use shift for future division */
				p = 8;			
				break;
			}
		}
	}
	if (ps > 3 || (((pwm_period * SYSTEM_TICK) >> p) == 0))
		return -22;
	
	T2CON = 0; 			/* Stop Timer2 and reset control reg */
	TMR2  = 0; 			/* Clear content of the count reg */
	IFS0bits.T2IF = 0;	 	/* Clear Timer2 interrupt flag */
	IEC0bits.T2IE = 0;		/* Disable Timer2 interrupt */
	T2CONbits.TCKPS = ps;		/* Set Timer2 pre-scaler reg */	
	timer2_prescaler = p;		/* Save computed pre-scaling factor */
	PR2 = ((pwm_period * SYSTEM_TICK) >> p) - 1; /* Set Timer2 period reg */
	T2CONbits.TON = 1;

	return RETIS_ERR_NONE;
}

COMPILER_INLINE void oc8_start(uint16_t init_ocr) 
{
	TRISDbits.TRISD7 = 0; 	/* Set OC8 as output */
	OC8R = init_ocr; 	/* Set the initial pw equivalent value */
	OC8RS = init_ocr; 	/* ... */
	OC8CON = 0x0006; 	/* Set OC8 modes: PWM, no fault check, Timer2 */
}

COMPILER_INLINE void oc7_start(uint16_t init_ocr) 
{
	TRISDbits.TRISD6 = 0; 	/* Set OC7 as output */
	OC7R = init_ocr; 	/* Set the initial pw equivalent value */
	OC7RS = init_ocr; 	/* ... */
	OC7CON = 0x0006; 	/* Set OC7 modes: PWM, no fault check, Timer2 */
}

COMPILER_INLINE void oc8_stop(void)
{
	OC8R = 0;
	OC8CON = 0;
}

COMPILER_INLINE void oc7_stop(void)
{
	OC7R = 0;
	OC7CON = 0;
}  

#define GET_OCR(stat, angle) \
	(PR2 - ((stat).ocr_at_zero + ((stat).ocr_unit * angle) - 1))

static int8_t parse_conf_data(const struct retis_servomotor_conf_data *info)
{
	uint32_t p;

	if (info->angle_unit != 1)
		return -RETIS_ERR_UNIMPLEMENTED;
	if (info->min_angle > 180 || info->min_angle < -180)
		return -RETIS_ERR_BAD_INIT_DATA;
	if (info->max_angle > 180 || info->max_angle < -180)
		return -RETIS_ERR_BAD_INIT_DATA;
	if (info->min_angle > info->max_angle)
		return -RETIS_ERR_BAD_INIT_DATA;
	if ((int32_t)((info->pw_per_unit * info->max_angle) + 
	    info->pw_at_zero) < 0)
		return -RETIS_ERR_BAD_INIT_DATA;
	p = (info->pw_per_unit * info->max_angle) + info->pw_at_zero;
	if (p > info->pwm_period)
		return -RETIS_ERR_BAD_INIT_DATA;
	if ((int32_t)((info->pw_per_unit * info->min_angle) + 
	    info->pw_at_zero) < 0)
		return -RETIS_ERR_BAD_INIT_DATA;
	if (p > info->pwm_period)
		return -RETIS_ERR_BAD_INIT_DATA;
	p = ((p * SYSTEM_TICK) >> timer2_prescaler) - 1;
	if (p == -1 || p == 0)
		return -RETIS_ERR_BAD_INIT_DATA;

	return RETIS_ERR_NONE;
}

int8_t retis_servomotor_init(uint8_t id, 
			     const struct retis_servomotor_conf_data *info)
{
	int8_t retv;

	switch (id) {
	case 0 :
		if ((int8_t )timer2_prescaler == -1) {
			retv = timer2_init(info->pwm_period);
			if (retv != RETIS_ERR_NONE)
				return retv;
			retv = parse_conf_data(info);
			if (retv != RETIS_ERR_NONE)
				return retv;
		}
		servo_stat[0].min_angle = info->min_angle;
		servo_stat[0].max_angle = info->max_angle;
		servo_stat[0].ocr_at_zero = (info->pw_at_zero * SYSTEM_TICK) >>
					    timer2_prescaler;
		servo_stat[0].ocr_unit = (info->pw_per_unit * SYSTEM_TICK) >>
					    timer2_prescaler;
		break;
	case 1 :
		if ((int8_t)timer2_prescaler == -1) {
			retv = timer2_init(info->pwm_period);
			if (retv != RETIS_ERR_NONE)
				return retv;
			retv = parse_conf_data(info);
			if (retv != RETIS_ERR_NONE)
				return retv;
		}
		servo_stat[1].min_angle = info->min_angle;
		servo_stat[1].max_angle = info->max_angle;
		servo_stat[1].ocr_at_zero = (info->pw_at_zero * SYSTEM_TICK) >>
					    timer2_prescaler;
		servo_stat[1].ocr_unit = (info->pw_per_unit * SYSTEM_TICK) >>
					    timer2_prescaler;
		break;
	default :
		return -RETIS_ERR_BAD_MODULE_ID;
	}

	return RETIS_ERR_NONE;
}

int8_t retis_servomotor_conf(uint8_t id, uint8_t attribute, void *value)
{
	uint8_t v;

	if (attribute != RETIS_SERVOMOTOR_POWER)
		return -RETIS_ERR_BAD_ATTRIBUTE;
	v = *((uint8_t *) value);
	switch (id) {
	case 0 :
		if (servo_stat[0].ocr_unit == 0) 
			return -RETIS_ERR_NOT_INITIALIZED;
		if (v == 0)
			oc8_stop();
		else if (v == 1)
			oc8_start(GET_OCR(servo_stat[1], 0));
		else
			return -RETIS_ERR_BAD_ATTRIBUTE_VALUE;
		break;
	case 1 :
		if (servo_stat[1].ocr_unit == 0) 
			return -RETIS_ERR_NOT_INITIALIZED;
		if (v == 0)
			oc7_stop();
		else if (v == 1)
			oc7_start(GET_OCR(servo_stat[1], 0));
		else
			return -RETIS_ERR_BAD_ATTRIBUTE_VALUE;
		break;
	default :
		return -RETIS_ERR_BAD_MODULE_ID;
	}

	return RETIS_ERR_NONE;
}

int8_t retis_servomotor_write(uint8_t id, int16_t value)
{
	switch (id) {
	case 0 :
		if (value < servo_stat[0].min_angle || 
		    value > servo_stat[0].max_angle)
			return -RETIS_ERR_BAD_VALUE;
		OC8RS = GET_OCR(servo_stat[0], value);
		break;
	case 1 :
		if (value < servo_stat[1].min_angle || 
		    value > servo_stat[1].max_angle)
			return -RETIS_ERR_BAD_VALUE;
		OC7RS = GET_OCR(servo_stat[1], value);
		break;
	default :
		return -RETIS_ERR_BAD_MODULE_ID;
	}

	return RETIS_ERR_NONE;
}

int8_t retis_servomotor_read(uint8_t id, int16_t *value)
{
	if (id >= RETIS_SERVOMOTOR_N)
		return -RETIS_ERR_BAD_MODULE_ID;
	return -RETIS_ERR_UNIMPLEMENTED;
}
